"""Tests for optional shape and type annotation features."""

import jax
import numpy as onp
import pytest
from jax import numpy as jnp
from typing_extensions import Annotated

import jax_dataclasses as jdc


@jdc.pytree_dataclass
class MnistStruct(jdc.EnforcedAnnotationsMixin):
    image: Annotated[
        onp.ndarray,
        (..., 28, 28),
        jnp.floating,
    ]
    label: Annotated[
        onp.ndarray,
        (..., 10),
        jnp.integer,
    ]


@jdc.pytree_dataclass
class MnistStructPartial(jdc.EnforcedAnnotationsMixin):
    image_shape_only: Annotated[
        onp.ndarray,
        (28, 28),  # Ellipsis will be appended automatically.
    ]
    label_dtype_only: Annotated[
        onp.ndarray,
        jnp.integer,
    ]


def test_valid() -> None:
    data = MnistStruct(
        image=onp.zeros((28, 28), dtype=onp.float32),
        label=onp.zeros((10,), dtype=onp.uint8),
    )
    assert data.get_batch_axes() == ()

    data = MnistStruct(
        image=onp.zeros((5, 28, 28), dtype=onp.float32),
        label=onp.zeros((5, 10), dtype=onp.uint8),
    )
    assert data.get_batch_axes() == (5,)

    data = MnistStruct(
        image=onp.zeros((5, 7, 28, 28), dtype=onp.float32),
        label=onp.zeros((5, 7, 10), dtype=onp.uint8),
    )
    assert data.get_batch_axes() == (5, 7)

    data_partial = MnistStructPartial(
        image_shape_only=onp.zeros((7, 28, 28), dtype=onp.float32),
        label_dtype_only=onp.zeros((70), dtype=onp.int32),
    )
    assert data_partial.get_batch_axes() == (7,)


def test_shape_mismatch() -> None:
    with pytest.raises(AssertionError):
        MnistStruct(
            image=onp.zeros((7, 32, 32), dtype=onp.float32),
            label=onp.zeros((7, 10), dtype=onp.uint8),
        )

    with pytest.raises(AssertionError):
        MnistStructPartial(
            image_shape_only=onp.zeros((7, 32, 32), dtype=onp.float32),
            label_dtype_only=onp.zeros((7, 10), dtype=onp.uint8),
        )


def test_batch_axis_mismatch() -> None:
    with pytest.raises(AssertionError):
        MnistStruct(
            image=onp.zeros((5, 7, 28, 28), dtype=onp.float32),
            label=onp.zeros((7, 10), dtype=onp.uint8),
        )


def test_dtype_mismatch() -> None:
    with pytest.raises(AssertionError):
        MnistStruct(
            image=onp.zeros((7, 28, 28), dtype=onp.uint8),
            label=onp.zeros((7, 10), dtype=onp.uint8),
        )

    with pytest.raises(AssertionError):
        MnistStructPartial(
            image_shape_only=onp.zeros((7, 28, 28), dtype=onp.float32),
            label_dtype_only=onp.zeros((7, 10), dtype=onp.float32),
        )


def test_nested() -> None:
    @jdc.pytree_dataclass
    class Parent(jdc.EnforcedAnnotationsMixin):
        x: Annotated[onp.ndarray, jnp.floating, ()]
        child: MnistStruct

    # OK
    assert Parent(
        x=onp.zeros((7,), dtype=onp.float32),
        child=MnistStruct(
            image=onp.zeros((7, 28, 28), dtype=onp.float32),
            label=onp.zeros((7, 10), dtype=onp.uint8),
        ),
    ).get_batch_axes() == (7,)

    # Batch axis mismatch
    with pytest.raises(AssertionError):
        Parent(
            x=onp.zeros((5,), dtype=onp.float32),
            child=MnistStruct(
                image=onp.zeros((7, 28, 28), dtype=onp.float32),
                label=onp.zeros((7, 10), dtype=onp.uint8),
            ),
        )

    # Type error
    with pytest.raises(AssertionError):
        Parent(
            x=onp.zeros((7,), dtype=onp.float32),
            child=MnistStruct(
                image=onp.zeros((7, 28, 28), dtype=onp.float32),
                label=onp.zeros((7, 10), dtype=onp.float32),
            ),
        )


def test_scalar() -> None:
    @jdc.pytree_dataclass
    class ScalarContainer(jdc.EnforcedAnnotationsMixin):
        scalar: Annotated[onp.ndarray, ()]  # () => scalar shape

    assert ScalarContainer(scalar=5.0).get_batch_axes() == ()  # type: ignore
    assert ScalarContainer(scalar=onp.zeros((5,))).get_batch_axes() == (5,)


def test_grad() -> None:
    @jdc.pytree_dataclass
    class Vector3(jdc.EnforcedAnnotationsMixin):
        parameters: Annotated[onp.ndarray, (3,)]

    # Make sure we can compute gradients wrt annotated dataclasses.
    grad = jax.grad(lambda x: jnp.sum(x.parameters))(Vector3(onp.zeros(3)))
    onp.testing.assert_allclose(grad.parameters, onp.ones((3,)))


def test_unannotated() -> None:
    @jdc.pytree_dataclass
    class Test(jdc.EnforcedAnnotationsMixin):
        a: onp.ndarray

    with pytest.raises(AssertionError):
        Test(onp.zeros((2, 1, 2, 3, 5, 7, 9))).get_batch_axes()


def test_middle_batch_axes() -> None:
    @jdc.pytree_dataclass
    class Test(jdc.EnforcedAnnotationsMixin):
        a: Annotated[onp.ndarray, (3, ..., 5, 7, 9)]

    test = Test(onp.zeros((3, 1, 2, 3, 5, 7, 9)))
    assert test.get_batch_axes() == (1, 2, 3)

    with pytest.raises(AssertionError):
        Test(onp.zeros((2, 1, 2, 3, 5, 7, 9)))
    with pytest.raises(AssertionError):
        Test(onp.zeros((3, 1, 2, 3, 5, 7)))


# This test currently breaks -- shape assertions on instantiation makes it impossible to
# compute some more complex Jacobians.
#
# Some options for fixing: adding some way to temporarily disable validation, or moving
# away from validation on instantiation to validation only when `.get_batch_axes()` is
# called. Either should be fairly straightforward, but this is fairly niche, produces (in my
# opinion) unintuitive Pytree structures, and is easy to work around, so marking as a
# no-fix for now.
#
# def test_jacobians() -> None:
#     @jdc.pytree_dataclass
#     class Vector3(ArrayAnnotationMixin):
#         parameters: Annotated[onp.ndarray, (3,)]
#
#     @jdc.pytree_dataclass
#     class Vector4(ArrayAnnotationMixin):
#         parameters: Annotated[onp.ndarray, (4,)]
#
#     def vec4_from_vec3(vec3: Vector3) -> Vector4:
#         return Vector4(onp.zeros((4,)))
#
#     def vec3_from_vec4(vec4: Vector4) -> Vector3:
#         return Vector3(onp.zeros((3,)))
#
#     jac = jax.jacfwd(vec4_from_vec3)(Vector3(onp.zeros((3,))))
#     jac = jax.jacfwd(vec3_from_vec4)(Vector4(onp.zeros((4,))))



from dataclasses import asdict, astuple, field, fields, is_dataclass, replace
from typing import TYPE_CHECKING

from ._copy_and_mutate import copy_and_mutate as copy_and_mutate

if TYPE_CHECKING:
    # Treat our JAX field and dataclass functions as their counterparts from the
    # standard dataclasses library during static analysis
    #
    # Tools like via mypy, jedi, etc generally rely on a lot of special, hardcoded
    # behavior for the standard dataclasses library; this lets us take advantage of all
    # of it.
    #
    # Note that mypy will not follow aliases, so `from dataclasses import dataclass` is
    # preferred over `dataclass = dataclasses.dataclass`.
    #
    # Dataclass transforms serve a similar purpose, but are currently only supported in
    # pyright and pylance.
    # https://github.com/microsoft/pyright/blob/master/specs/dataclass_transforms.md
    # `static_field()` is deprecated, but not a lot of code to support, so leaving it
    # for now...
    from dataclasses import dataclass as pytree_dataclass
else:
    from ._dataclasses import pytree_dataclass  # noqa
    from ._dataclasses import deprecated_static_field as static_field  # noqa

from ._dataclasses import Static
from ._enforced_annotations import EnforcedAnnotationsMixin
from ._jit import jit

__all__ = [
    "asdict",
    "astuple",
    "field",
    "fields",
    "is_dataclass",
    "replace",
    "copy_and_mutate",
    "pytree_dataclass",
    "Static",
    "EnforcedAnnotationsMixin",
    "jit",
]



"""Tests for serialization using `flax.serialization`."""

import flax
import numpy as onp
from jax import tree_util

import jax_dataclasses as jdc


def _assert_pytree_allclose(x, y) -> None:
    tree_util.tree_map(
        lambda *arrays: onp.testing.assert_allclose(arrays[0], arrays[1]), x, y
    )


def test_serialization() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: int
        field2: int
        field3: jdc.Static[bool]

    obj = A(field1=5, field2=3, field3=True)

    _assert_pytree_allclose(
        obj,
        flax.serialization.from_bytes(obj, flax.serialization.to_bytes(obj)),
    )



"""Tests for standard jdc.pytree_dataclass features. Initialization, flattening, unflattening,
static fields, etc.
"""

from __future__ import annotations

from typing import Generic, TypeVar

import jax
import numpy as onp
import pytest
from jax import tree_util

import jax_dataclasses as jdc


def _assert_pytree_allclose(x, y):
    tree_util.tree_map(
        lambda *arrays: onp.testing.assert_allclose(arrays[0], arrays[1]), x, y
    )


def test_init() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: int
        field2: int

    assert A(field1=5, field2=3) == A(5, 3)

    with pytest.raises(TypeError):
        # Not enough arguments
        A(field1=5)  # type: ignore


def test_default_arg() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: int
        field2: int = 3

    assert A(field1=5, field2=3) == A(5, 3) == A(field1=5) == A(5)


def test_flatten() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: float
        field2: float

    @jax.jit
    def jitted_sum(obj: A) -> float:
        return obj.field1 + obj.field2

    _assert_pytree_allclose(jitted_sum(A(5.0, 3.0)), 8.0)


def test_unflatten() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: float
        field2: float

    @jax.jit
    def construct_A(a: float) -> A:
        return A(field1=a, field2=a * 2.0)

    _assert_pytree_allclose(A(1.0, 2.0), construct_A(1.0))


def test_static_field() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: float
        field2: float
        field3: jdc.Static[bool]

    @jax.jit
    def jitted_op(obj: A) -> float:
        if obj.field3:
            return obj.field1 + obj.field2
        else:
            return obj.field1 - obj.field2

    with pytest.raises(ValueError):
        # Cannot map over pytrees with different treedefs
        _assert_pytree_allclose(A(1.0, 2.0, False), A(1.0, 2.0, True))

    _assert_pytree_allclose(jitted_op(A(5.0, 3.0, True)), 8.0)
    _assert_pytree_allclose(jitted_op(A(5.0, 3.0, False)), 2.0)


def test_static_field_deprecated() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: float
        field2: float
        field3: bool = jdc.static_field()  # type: ignore

    @jax.jit
    def jitted_op(obj: A) -> float:
        if obj.field3:
            return obj.field1 + obj.field2
        else:
            return obj.field1 - obj.field2

    with pytest.raises(ValueError):
        # Cannot map over pytrees with different treedefs
        _assert_pytree_allclose(A(1.0, 2.0, False), A(1.0, 2.0, True))

    _assert_pytree_allclose(jitted_op(A(5.0, 3.0, True)), 8.0)
    _assert_pytree_allclose(jitted_op(A(5.0, 3.0, False)), 2.0)


def test_no_init() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: float
        field2: float = jdc.field()
        field3: jdc.Static[bool] = jdc.field(init=False)

        def __post_init__(self):
            object.__setattr__(self, "field3", False)

    @jax.jit
    def construct_A(a: float) -> A:
        return A(field1=a, field2=a * 2.0)

    assert construct_A(5.0).field3 is False


def test_static_field_forward_ref() -> None:
    @jdc.pytree_dataclass
    class A:
        field1: float
        field2: float
        field3: jdc.Static[Container[bool]]

    T = TypeVar("T")

    @jdc.pytree_dataclass
    class Container(Generic[T]):
        x: T

    @jax.jit
    def jitted_op(obj: A) -> float:
        if obj.field3.x:
            return obj.field1 + obj.field2
        else:
            return obj.field1 - obj.field2

    with pytest.raises(ValueError):
        # Cannot map over pytrees with different treedefs
        _assert_pytree_allclose(
            A(1.0, 2.0, Container(False)), A(1.0, 2.0, Container(True))
        )

    _assert_pytree_allclose(jitted_op(A(5.0, 3.0, Container(True))), 8.0)
    _assert_pytree_allclose(jitted_op(A(5.0, 3.0, Container(False))), 2.0)



# mypy: ignore-errors
#
# PEP 695 generics aren't yet supported in mypy.

from __future__ import annotations

from typing import Generic, TypeVarTuple

import jax_dataclasses as jdc

Ts = TypeVarTuple("Ts")


@jdc.pytree_dataclass
class Args(Generic[*Ts]):
    @staticmethod
    @jdc.jit
    def make(args: jdc.Static[tuple[*Ts]]) -> tuple[Args[*Ts], tuple[*Ts]]:
        return Args(), args


def test0() -> None:
    assert Args.make((1, 2, 3))[1] == (1, 2, 3)


@jdc.pytree_dataclass
class Args2[*T]:
    @staticmethod
    @jdc.jit
    def make[*T_](args: jdc.Static[tuple[*T_]]) -> tuple[Args2[*T_], tuple[*T_]]:
        return Args2(), args


def test1() -> None:
    assert Args2.make((1, 2, 3))[1] == (1, 2, 3)



import dataclasses
from typing import Any, Dict, List

import numpy as onp
import pytest

import jax_dataclasses as jdc


def test_copy_and_mutate() -> None:
    # frozen=True should do nothing
    @jdc.pytree_dataclass(frozen=True)
    class Foo:
        array: Any

    @jdc.pytree_dataclass
    class Bar:
        children: List[Foo]
        array: Any
        array_unchanged: onp.ndarray

    obj = Bar(
        children=[Foo(array=onp.zeros(3))],
        array=onp.ones(3),
        array_unchanged=onp.ones(3),
    )

    # Registered dataclasses are initially immutable
    with pytest.raises(dataclasses.FrozenInstanceError):
        obj.array = onp.zeros(3)

    # But we can use a context that copies a dataclass and temporarily makes the copy
    # mutable:
    with jdc.copy_and_mutate(obj) as obj:
        # Updates can then very easily be applied!
        obj.array = onp.zeros(3)
        obj.children[0].array = onp.ones(3)  # type: ignore

        # Shapes can be validated...
        with pytest.raises(AssertionError):
            obj.children[0].array = onp.ones(1)  # type: ignore

        # As well as dtypes
        with pytest.raises(AssertionError):
            obj.children[0].array = onp.ones(3, dtype=onp.int32)  # type: ignore

    # Validation can also be disabled
    with jdc.copy_and_mutate(obj, validate=False) as obj:
        obj.children[0].array = onp.ones(1)  # type: ignore
        obj.children[0].array = onp.ones(3)  # type: ignore

    # Outside of the replace context, the copied object becomes immutable again:
    with pytest.raises(dataclasses.FrozenInstanceError):
        obj.array = onp.zeros(3)
    with pytest.raises(dataclasses.FrozenInstanceError):
        obj.children[0].array = onp.ones(3)  # type: ignore

    onp.testing.assert_allclose(obj.array, onp.zeros(3))
    onp.testing.assert_allclose(obj.array_unchanged, onp.ones(3))
    onp.testing.assert_allclose(obj.children[0].array, onp.ones(3))


def test_copy_and_mutate_static() -> None:
    @dataclasses.dataclass
    class Inner:
        a: int
        b: int

    @jdc.pytree_dataclass
    class Foo:
        arrays: Dict[str, onp.ndarray]
        child: jdc.Static[Inner]

    obj = Foo(arrays={"x": onp.ones(3)}, child=Inner(1, 2))

    # Registered dataclasses are initially immutable
    with pytest.raises(dataclasses.FrozenInstanceError):
        obj.child = Inner(5, 6)

    assert obj.child == Inner(1, 2)

    # But can be copied and mutated in a special context
    with jdc.copy_and_mutate(obj) as obj_updated:
        obj_updated.child = Inner(5, 6)

    assert obj.child == Inner(1, 2)
    assert obj_updated.child == Inner(5, 6)



from __future__ import annotations

from typing import Tuple

from jax import numpy as jnp

import jax_dataclasses as jdc


def test_cycle() -> None:
    @jdc.pytree_dataclass
    class TreeNode:
        content: jnp.ndarray
        children: Tuple[TreeNode, ...]

    a = TreeNode(content=jnp.zeros(3), children=())
    b = TreeNode(content=jnp.ones(3), children=(a,))

    # Not a cycle. OK!
    with jdc.copy_and_mutate(b, validate=False) as b:
        b.children = (a, a)

    # Cycle. Ideally this should raise an error, but the way duplicate nodes in pytrees
    # are possible makes robust cycle detection under linear space/time constraints a
    # hassle. So instead we currently do nothing.
    with jdc.copy_and_mutate(b, validate=False) as b:
        b.children[0].children = (b,)



import inspect
from typing import Any, Callable, Optional, Sequence, TypeVar, Union, cast, overload

import jax
from jaxlib import xla_client as xc

from ._dataclasses import JDC_STATIC_MARKER
from ._get_type_hints import get_type_hints_partial

CallableType = TypeVar("CallableType", bound=Callable)


@overload
def jit(
    fun: CallableType,
    *,
    device: Optional[xc.Device] = None,
    backend: Optional[str] = None,
    donate_argnums: Union[int, Sequence[int]] = (),
    inline: bool = False,
    keep_unused: bool = False,
    abstracted_axes: Optional[Any] = None,
) -> CallableType: ...


@overload
def jit(
    fun: None = None,
    *,
    device: Optional[xc.Device] = None,
    backend: Optional[str] = None,
    donate_argnums: Union[int, Sequence[int]] = (),
    inline: bool = False,
    keep_unused: bool = False,
    abstracted_axes: Optional[Any] = None,
) -> Callable[[CallableType], CallableType]: ...


def jit(
    fun: Optional[CallableType] = None,
    *,
    device: Optional[xc.Device] = None,
    backend: Optional[str] = None,
    donate_argnums: Union[int, Sequence[int]] = (),
    inline: bool = False,
    keep_unused: bool = False,
    abstracted_axes: Optional[Any] = None,
) -> Union[CallableType, Callable[[CallableType], CallableType]]:
    """Light wrapper around `jax.jit`, with usability and type checking improvements.

    Three differences:
        - We remove the `static_argnums` and `static_argnames` parameters. Instead,
          static arguments can be specified in type annotations with
          `jax_dataclasses.Static[]`.
        - Instead of `jax.stages.Wrapped`, the return callable type is annotated to
          match the input callable type. This will improve autocomplete and type
          checking in most situations.
        - Similar to `@dataclasses.dataclass`, return a decorator if `fun` isn't passed
          in. This is convenient for avoiding `@functools.partial()`.
    """

    def wrap(fun: CallableType) -> CallableType:
        signature = inspect.signature(fun)

        # Mark any inputs annotated with jax_dataclasses.Static[] as static.
        static_argnums = []
        static_argnames = []
        hint_from_name = get_type_hints_partial(fun, include_extras=True)
        for i, param in enumerate(signature.parameters.values()):
            name = param.name
            if name not in hint_from_name:
                continue
            hint = hint_from_name[name]
            if hasattr(hint, "__metadata__") and JDC_STATIC_MARKER in hint.__metadata__:
                if param.kind is param.POSITIONAL_ONLY:
                    static_argnums.append(i)
                else:
                    static_argnames.append(name)

        return cast(
            CallableType,
            jax.jit(
                fun,
                static_argnums=static_argnums if len(static_argnums) > 0 else None,
                static_argnames=static_argnames if len(static_argnames) > 0 else None,
                device=device,
                backend=backend,
                donate_argnums=donate_argnums,
                inline=inline,
                keep_unused=keep_unused,
                abstracted_axes=abstracted_axes,
            ),
        )

    if fun is None:
        return wrap
    else:
        return wrap(fun)



from __future__ import annotations

import collections
import functools
import sys
import types
from types import MethodDescriptorType, MethodWrapperType, WrapperDescriptorType
from typing import Any, Dict, Type


class _UnresolvableForwardReference:
    def __class_getitem__(cls, item) -> Type[_UnresolvableForwardReference]:
        """__getitem__ passthrough, for supporting generics."""
        return _UnresolvableForwardReference


_allowed_types = (
    types.FunctionType,
    types.BuiltinFunctionType,
    types.MethodType,
    types.ModuleType,
    WrapperDescriptorType,
    MethodWrapperType,
    MethodDescriptorType,
)


@functools.lru_cache(maxsize=128)
def get_type_hints_partial(obj, include_extras=False) -> Dict[str, Any]:
    """Adapted from typing.get_type_hints(), but aimed at suppressing errors from not
    (yet) resolvable forward references.

    This function should only be used to search for fields that are annotated with
    `jdc.Static[]`.

    For example:

        @jdc.pytree_dataclass
        class A:
            x: B
            y: jdc.Static[bool]

        @jdc.pytree_dataclass
        class B:
            x: jnp.ndarray

    Note that the type annotations of `A` need to be parsed by the `pytree_dataclass`
    decorator in order to register the static field, but `B` is not yet defined when the
    decorator is run. We don't actually care about the details of the `B` annotation, so
    we replace it in our annotation dictionary with a dummy value.

    Differences:
        1. `include_extras` must be True.
        2. Only supports types.
        3. Doesn't throw an error when a name is not found. Instead, replaces the value
           with `_UnresolvableForwardReference`.
    """
    assert include_extras

    # Replace any unresolvable names with _UnresolvableForwardReference.
    base_globals: Dict[str, Any] = collections.defaultdict(
        lambda: _UnresolvableForwardReference
    )
    base_globals.update(__builtins__)  # type: ignore

    # Classes require a special treatment.
    if isinstance(obj, type):
        hints = {}
        for base in reversed(obj.__mro__):
            ann = base.__dict__.get("__annotations__", {})
            if len(ann) == 0:
                continue

            base_globals.update(sys.modules[base.__module__].__dict__)

            for name, value in ann.items():
                if value is None:
                    value = type(None)
                if isinstance(value, str):
                    # The * replace is a hack for variadic generic support.
                    value = value.replace("*", "")
                    value = eval(value, base_globals)
                hints[name] = value
        return hints

    nsobj = obj

    # Find globalns for the unwrapped object.
    while hasattr(nsobj, "__wrapped__"):
        nsobj = nsobj.__wrapped__
    base_globals.update(getattr(nsobj, "__globals__", {}))

    hints = getattr(obj, "__annotations__", None)  # type: ignore
    if hints is None:
        # Return empty annotations for something that _could_ have them.
        if isinstance(obj, _allowed_types):
            return {}
        else:
            raise TypeError(
                "{!r} is not a module, class, method, or function.".format(obj)
            )
    hints = dict(hints)
    for name, value in hints.items():
        if value is None:
            value = type(None)
        if isinstance(value, str):
            # The * replace is a hack for variadic generic support.
            value = value.replace("*", "")
            value = eval(value, base_globals)
        hints[name] = value
    return hints



import jax
import pytest
from jax import numpy as jnp
from typing_extensions import Annotated

import jax_dataclasses as jdc


@jdc.pytree_dataclass
class Node(jdc.EnforcedAnnotationsMixin):
    a: Annotated[jnp.ndarray, (5,), jnp.floating]


def test_vmap():
    with pytest.raises(AssertionError):
        jax.jit(jax.vmap(lambda *unused: None))(Node(jnp.zeros((5,))))

    jax.jit(jax.vmap(lambda *unused: None))(Node(jnp.zeros((5, 5))))
    jax.jit(jax.vmap(lambda *unused: None, in_axes=(None, None, 0, 0)))(
        Node(jnp.zeros(5)),
        Node(jnp.zeros(5)),
        jnp.zeros((5, 100)),
        jnp.zeros((5, 100)),
    )



from __future__ import annotations

import dataclasses

import jax
import pytest
from jax import numpy as jnp

import jax_dataclasses as jdc


def test_jit_0():
    def func(x: int, y: int) -> jax.Array:
        assert isinstance(x, int)
        assert isinstance(y, int)
        return jnp.full(shape=(x,), fill_value=y)

    assert func(3, 4).shape == (3,)


def test_jit_1():
    @jdc.jit
    def func(x: int, y: int) -> jax.Array:
        assert not isinstance(x, int)
        assert not isinstance(y, int)
        return jnp.full(shape=(x,), fill_value=y)

    with pytest.raises(TypeError):
        assert func(3, 4).shape == (3,)


def test_jit_2():
    @jdc.jit
    def func(x: jdc.Static[int], y: int) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        return jnp.full(shape=(x,), fill_value=y)

    assert func(3, 4).shape == (3,)


def test_jit_3():
    @jdc.jit
    def func(x: jdc.Static[int], /, y: int) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        return jnp.full(shape=(x,), fill_value=y)

    assert func(3, 4).shape == (3,)


def test_jit_4():
    @jdc.jit
    def func(*, x: jdc.Static[int], y: int) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        return jnp.full(shape=(x,), fill_value=y)

    assert func(x=3, y=4).shape == (3,)


def test_jit_5():
    @jdc.jit
    def func(x: jdc.Static[int], y: int, /, *, z: jdc.Static[int]) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        assert isinstance(z, int)
        return jnp.full(shape=(x + z,), fill_value=y)

    assert func(2, 4, z=1).shape == (3,)


def test_jit_6():
    @jdc.jit
    def func(x: jdc.Static[int], y: int, *, z: jdc.Static[int]) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        assert isinstance(z, int)
        return jnp.full(shape=(x + z,), fill_value=y)

    assert func(2, 4, z=1).shape == (3,)


def test_jit_7():
    @jdc.jit
    def func(x: jdc.Static[int], y: int, z: jdc.Static[int], /) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        assert isinstance(z, int)
        return jnp.full(shape=(x + z,), fill_value=y)

    assert func(2, 4, 1).shape == (3,)


def test_jit_no_annotation():
    @jdc.jit
    def func(x: jdc.Static[int], y, z: jdc.Static[int], /) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        assert isinstance(z, int)
        return jnp.full(shape=(x + z,), fill_value=y)

    assert func(2, 4, 1).shape == (3,)


def test_jit_donate_buffer():
    @jdc.jit(donate_argnums=(1,))
    def func(x: jdc.Static[int], y: int, z: jdc.Static[int], /) -> jax.Array:
        assert isinstance(x, int)
        assert not isinstance(y, int)
        assert isinstance(z, int)
        out = jnp.full(shape=(x + z,), fill_value=y)

        # Shape matches `y`, so we should be able to reuse the donated buffer.
        return jnp.sum(out)

    assert func(2, 4, 1).shape == ()


def test_jit_forward_ref():
    @jdc.jit
    def func(xz: jdc.Static[SomeConfig], y: int, /) -> jax.Array:
        assert not isinstance(y, int)
        return jnp.full(shape=(xz.x + xz.z,), fill_value=y)

    assert func(SomeConfig(2, 1), 4).shape == (3,)


def test_jit_lambda():
    assert jdc.jit(lambda x, y: x + y)(jnp.zeros(3), jnp.ones(3)).shape == (3,)


@dataclasses.dataclass(frozen=True)
class SomeConfig:
    x: int
    z: int
